#!/usr/bin/env python3
"""Demonstrate a basic program that generates OpenTelemetry tracing."""
import logging
import os
import random
import time

import opentelemetry

from form_observability import (
    ContextAwareTracer,
    OtelSpanEventHandler,
    configure,
    ctx,
)


_trace = ContextAwareTracer(__name__)
_log = logging.getLogger(__name__)


def _helper_that_sometimes_catches_exceptions_internally(test_id: str) -> None:
    sleep_s = random.random()
    err_f = random.random()
    # Set some attributes that will be added to any spans and events in this context.
    with ctx.set(
        {
            "test_size": len(test_id),
            "sleep_s": sleep_s,
            "err_f": err_f,
        },
        # Also set these attributes on the active span.
        update_current_span=True,
    ):
        try:
            time.sleep(sleep_s)
            if err_f < 0.1:
                raise RuntimeError("Demo error to show exception treatment.")
            # This success event will have the test_size attribute set on it from
            # the context above, as well as the test_id attribute set in main().
            _trace.add_event("success")
            return True
        except:
            _trace.add_event("failure")
            _log.exception(f"Error during processing of {test_id=}.")
            return False


@_trace.traced  # Open a span for this function call. In this case it's the root span.
def main():
    _log.info("Preparing to process tests.")
    num_errors = 0
    for test_id in ("one", "two", "three", "four", "five", "six"):
        with _trace.start_as_current_span("work", attributes={"test_id": test_id}):
            if _helper_that_sometimes_catches_exceptions_internally(test_id):
                num_errors += 1
    opentelemetry.trace.get_current_span().set_attributes({"num_errors": num_errors})
    _log.info("Demo processing complete.")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    header_env = "OTEL_EXPORTER_OTLP_HEADERS"
    endpoint_env = "OTEL_EXPORTER_OTLP_ENDPOINT"
    if header_env not in os.environ:
        _log.error(f"To upload traces, you must set {header_env}, see README.md.")
    # Set up a simple span exporter with no sampling, suitable for data pipelines.
    configure(
        "form_observability_example",  # demo dataset name
        otlp_headers=os.environ.get(header_env),
        otlp_endpoint=os.environ.get(endpoint_env, "https://api.honeycomb.io"),
    )
    _log.addHandler(OtelSpanEventHandler())  # Add span events for log messages.
    main()



from collections import namedtuple
import time

from dagster import DagsterEventType

from form_observability.dagster_otel.util import get_run_id_and_ancestors


TimingInfo = namedtuple("TimingInfo", ["start_s", "end_s", "dt_s"])


def get_timing_since_step_started(
    context,
    step_key: str,
    seconds_fn=time.time,
) -> TimingInfo:
    """Returns timing information relative to the provided Dagster step key.

    :return: A tuple of timing information: start_s (when the named step started in
        epoch seconds), end_s (now), and dt_s (the time delta in seconds).
    """
    t = seconds_fn()
    run_ids = get_run_id_and_ancestors(context)
    all_searched_logs = []
    for run_id in run_ids:
        logs = context.instance.all_logs(run_id, DagsterEventType.STEP_START)
        all_searched_logs += logs
        for event_log_entry in logs:
            if event_log_entry.dagster_event.step_handle.key == step_key:
                timestamp_s = event_log_entry.timestamp
                return TimingInfo(timestamp_s, t, t - timestamp_s)
    raise RuntimeError(
        f"Cannot add elapsed time, step key {step_key} not found in STEP_START"
        f" events: {[e.dagster_event.step_handle.key for e in all_searched_logs]}."
        f" Searched run IDs: {run_ids}."
    )



import logging
import traceback

import opentelemetry
from opentelemetry.trace import Status, StatusCode
from opentelemetry.semconv.trace import SpanAttributes

from form_observability.context_aware import ctx
from form_observability.otel_value import EventAttrKey, EventAttrValue


class OtelSpanEventHandler(logging.Handler):
    """Attach log messages as events to the current span. Noop if no current span.

    Also sets the span error status if the log is ERROR or above.

    Following Honeycomb advice:
    https://www.honeycomb.io/blog/uniting-tracing-logs-open-telemetry-span-events/
    """

    def emit(self, record: logging.LogRecord) -> None:
        span = opentelemetry.trace.get_current_span()
        # Records logged when no trace is active will have no trace ID; in practice
        # get_current_span seems to always return a span but when it's not real it
        # has 0 values for IDs. This will happen for Dagster framework logs before
        # and after op execution, b/c span context is only activated during an otel_op.
        if not (span and span.get_span_context().trace_id):
            return

        # Where possible, follow the semantic conventions of span attributes. See
        # https://github.com/open-telemetry/opentelemetry-python/blob/main/opentelemetry-semantic-conventions/src/opentelemetry/semconv/trace/__init__.py
        attributes = {
            EventAttrKey.TYPE: EventAttrValue.LOG_MESSAGE,
            "logger.name": record.name,
            "level": record.levelname,
            SpanAttributes.CODE_FILEPATH: record.pathname,
            "code.filename": record.filename,
            SpanAttributes.CODE_LINENO: record.lineno,
            SpanAttributes.CODE_NAMESPACE: record.module,
            SpanAttributes.CODE_FUNCTION: record.funcName,
            # These are conceptually resources, and could be set on the span when it
            # is started instead of logged repeatedly with messages.
            "process.pid": record.process,
            "process.name": record.processName,
            SpanAttributes.THREAD_ID: record.thread,
            SpanAttributes.THREAD_NAME: record.threadName,
        }

        if record.exc_info is not None:
            exc_type, exc_value, tb = record.exc_info
            # Record similar attributes as span.record_exception, but don't use it
            # to avoid having two events.
            attributes.update(
                {
                    SpanAttributes.EXCEPTION_TYPE: exc_value.__class__.__name__,
                    SpanAttributes.EXCEPTION_MESSAGE: str(exc_value),
                    SpanAttributes.EXCEPTION_STACKTRACE: "".join(
                        traceback.format_exception(*record.exc_info)
                    ),
                    SpanAttributes.EXCEPTION_ESCAPED: False,
                }
            )

        if record.levelno >= logging.ERROR:
            # Set the span's error status, which triggers special UI treatment. Also
            # set the log message as the span's status_message, but note that if there
            # are additional error logs the span will only show the latest (but events
            # will exist for all of the error logs).
            span.set_status(Status(StatusCode.ERROR, record.getMessage()))
            # Setting error status on a span adds the "error" attribute. Mimic that
            # for the event too.
            attributes["error"] = True

        attributes.update(ctx.current_ctx)  # TODO if updating to Python 3.9, use | .
        span.add_event(
            # Use the log message as the event name, for Honeycomb UI visibility.
            record.getMessage(),
            attributes=attributes,
        )



from unittest.mock import MagicMock, patch
import pytest

import opentelemetry
from opentelemetry import context
from opentelemetry.trace.span import Span

from form_observability.context_aware import ContextAwareTracer, ctx
from form_observability.otel_value import EventAttrKey, EventAttrValue


class _TestAttrValue(EventAttrValue):
    UNITTEST = "unit_test"


@patch.object(opentelemetry.trace, "get_current_span")
def test_add_duration_event(mock_get_current_span):
    mock_span = MagicMock(spec=Span)
    mock_get_current_span.return_value = mock_span
    dt = 12.5

    tracer = ContextAwareTracer("unittest")
    tracer.add_duration_event(_TestAttrValue.UNITTEST, dt, {"test": True})

    mock_span.add_event.assert_called_once()
    mock_call = mock_span.add_event.call_args_list[0]
    assert mock_call.args[0] == _TestAttrValue.UNITTEST
    event_attrs = mock_call.kwargs.get("attributes")
    assert event_attrs.get(EventAttrKey.TYPE) == EventAttrValue.DURATION
    assert event_attrs.get(EventAttrKey.DURATION_SECONDS) == dt
    assert event_attrs.get(EventAttrKey.DURATION_NAME) == _TestAttrValue.UNITTEST
    assert event_attrs.get("test") == True


def test_context():
    assert ctx.get("a") is None

    with ctx.set({"a": 1}), ctx.set({"b": 2}):
        assert ctx.get("a") == 1
        assert ctx.get("b") == 2

        with ctx.set({"a": 2}):
            assert ctx.get("a") == 2
            assert ctx.get("b") == 2

        assert ctx.get("a") == 1
        assert ctx.get("b") == 2

    assert ctx.get("a") is None
    assert ctx.get("b") is None



from collections import namedtuple
from functools import wraps
import inspect
import resource
import platform
import time
from typing import Optional, Dict, Any

import dagster
from dagster import AssetKey
from dagster import AssetMaterialization
from dagster import DagsterEventType
from dagster import Output
from dagster import op
import opentelemetry
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from form_observability.context_aware import ContextAwareTracer, ctx
from form_observability.otel_value import SpanName
from form_observability.dagster_otel.util import get_run_id_and_ancestors


#: A resource that configures OpenTelemetry exporting should be registered under
#: this key. For example:
#:
#: @dagster.resource
#: def configure_observability_resource():
#:     """Run observability configuration setup as a Dagster resource.
#:
#:     This ensures span processing is set up before each op runs, because resources are
#:     initialized once per execution context. So this will be called by Dagster once in
#:     each Python interpreter.
#:     """
#:     pipeline_setup.configure(..)
CONFIGURE_OBSERVABILITY_RESOURCE_KEY = "configure_observability"


_tracer = ContextAwareTracer(__name__)


def publish_current_trace_context(context: dagster.OpExecutionContext) -> None:
    """Stores the current trace context, to be picked up by @otel_op.

    Should be called exactly once per graph.

    The context published here is automatically picked up by @otel_op
    using _set_trace_context.

    The trace context is published as JSON metadata on an AssetMaterializationEvent,
    see Dagster event docs:
    https://docs.dagster.io/concepts/ops-jobs-graphs/op-events#op-events-and-exceptions
    This is just using Dagster events as a shared db that does not require further
    setup, but trace contexts could instead be stored in any other custom store. See
    https://github.com/dagster-io/dagster/issues/7581 which requests a way to pipe
    one op's output to all downstream ops.
    """
    span = opentelemetry.trace.get_current_span()
    if not span:
        raise RuntimeError("No active span, cannot publish trace context.")
    carrier = {}
    TraceContextTextMapPropagator().inject(carrier)

    # If this is in a subgraph, use the subgraph name as the trace key. Otherwise,
    # it's not in a subgraph, use the root span name.
    # Dagster's internal convention is to use dotted namespaces for subgraphs.
    op_path = context.solid_handle.path
    trace_key = SpanName.ROOT if len(op_path) <= 1 else ".".join(op_path[:-1])

    context.log_event(
        AssetMaterialization(
            asset_key=AssetKey(trace_key),
            description="OpenTelemetry trace context.",
            metadata={
                "trace_context": carrier,
            },
        ),
    )


def _find_trace_context(context):
    # Try to find a trace context for this subgraph, but fall back to parent graphs
    # up to the root.
    trace_keys = []
    remaining_path = context.solid_handle.path[:-1]
    while remaining_path:
        key = ".".join(remaining_path)
        remaining_path.pop()
        trace_keys.append(key)
    trace_keys.append(SpanName.ROOT)

    run_ids = get_run_id_and_ancestors(context)
    all_searched_logs = []
    for run_id in run_ids:
        logs = context.instance.all_logs(run_id, DagsterEventType.ASSET_MATERIALIZATION)
        all_searched_logs += logs

        for trace_key in trace_keys:
            for event_log_entry in logs:
                materialization = (
                    event_log_entry.dagster_event.event_specific_data.materialization
                )
                event_asset_key = materialization.asset_key.path[0]
                if event_asset_key == trace_key:
                    return materialization.metadata_entries[0].entry_data.data

    raise RuntimeError(
        f"Could not find trace context, searched for any of {trace_keys} in"
        f" {len(all_searched_logs)} events from run IDs {run_ids}."
    )


def _set_trace_context(trace_context):
    """Sets the current trace context from the given propagated trace context carrier.

    See Python trace context propagation docs:
    https://opentelemetry.io/docs/instrumentation/python/cookbook/#manually-setting-span-context
    and conceptual docs:
    https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/context/api-propagators.md
    """
    ctx = TraceContextTextMapPropagator().extract(trace_context)
    opentelemetry.context.attach(ctx)


def otel_op(
    *decorator_args,
    span_name: Optional[str] = None,
    on_exception_return: Optional[Output] = None,
    ctx_attributes: Dict[str, Any] = {},
    **decorator_kwargs,
):
    """A wrapper for @op which automatically activates OpenTelemetry trace contexts.

    This requires the configure_observability resource key for every @otel_op, to
    ensure span processing is set up. (In tests, provide a noop resource.)

    This looks for the published root trace context or any appropriate subgraph context,
    and activates it.

    Then it also starts a span named for the current function, or the provided name.

    :param span_name: A specific span name to use, for ease of viewing in observability
        services. If not given, the span name falls back to (a) any explicit Dagster
        @op name, or (b) the decorated function's name.
    :param on_exception_return: If the @op raises an exception, swallow it and yield
        this value from the op. This should be used for per-test processing steps which
        should not halt overall graph execution. Typically Output(None, "result").
    :param ctx_attributes: Span / context attributes for this op.
    """
    op_name = decorator_kwargs.get("name")

    rkeys = "required_resource_keys"
    if rkeys not in decorator_kwargs:
        decorator_kwargs[rkeys] = set()
    decorator_kwargs[rkeys].add(CONFIGURE_OBSERVABILITY_RESOURCE_KEY)

    def wrapper(func):
        @op(*decorator_args, **decorator_kwargs)
        @wraps(func)
        # If the wrapped op is missing a context argument, this will get an error like
        # "raw_op() missing 1 required positional argument: 'context'". But we need
        # the context below, and want Dagster to pass args according to the wrapped
        # functions args so just require all @otel_ops to take context.
        def raw_op(context, *func_args, **func_kwargs):
            # Check for possible/occasional OpenTelemetry shutdown hangs.
            # https://github.com/open-telemetry/opentelemetry-python/issues/2284

            trace_context = _find_trace_context(context)
            _set_trace_context(trace_context)
            name = span_name or op_name or func.__name__
            with ctx.set(ctx_attributes), _tracer.start_as_current_span(name) as span:
                try:
                    ret = func(context, *func_args, **func_kwargs)
                    if inspect.isgenerator(ret):
                        # Yield from the generator here rather than just returning it so
                        # that the generator executes within the context manager.
                        yield from ret
                    else:
                        yield Output(ret, "result")
                except Exception as e:
                    if on_exception_return:
                        # Logging will indirectly record a span exception.
                        context.log.exception(str(e))
                        yield on_exception_return
                    else:
                        # Span context managers automatically record exceptions exiting
                        # their context.
                        raise
                finally:
                    # Record the maximum resident set size (peak RAM usage) of this
                    # process. For a Python multiprocessing executor (or executor using
                    # a container per op) this will indicate the op's max memory usage.
                    maxrss = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                    maxrss_b = (
                        maxrss
                        * {
                            # Linux documents maxrss units as kilobytes.
                            "linux": 1000,
                            # MacOS X doesn't document units but they seem to be bytes,
                            # for example a fresh Python session might report 10989568.
                            "darwin": 1,
                        }[platform.system().lower()]
                    )
                    span.set_attributes(
                        {
                            "process.resource.self.maxrss_b": maxrss_b,
                        }
                    )

        return raw_op

    return wrapper



"""OpenTelemetry tracing setup.

Example usage for Honeycomb:

pipeline_setup.configure(
    "my-dataset",
    "x-honeycomb-team=<API key for desired environment>",
)
"""
import os

import opentelemetry
from opentelemetry.instrumentation.requests import RequestsInstrumentor

# Initialize tracing and an exporter that can send data to Honeycomb, New Relic, LightStep
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry.sdk.trace import TracerProvider, SpanLimits


def configure(
    service_name: str,
    otlp_headers: str,
    otlp_endpoint: str = "https://api.honeycomb.io",
    instrument_requests: bool = False,
) -> None:
    """Set up a trace provider and processor, so tracing data gets exported.

    This sets options which are tuned for a multithreaded data pipeline which should
    export all data (without sampling) via OTLP (not a collector).

    This should be called once before program execution within each Python interpreter.
    For example, this can be invoked as a Dagster resource for each @op.

    Optionally auto-instruments the Python requests library. (SqlAlchemy instrumentation
    is added to each engine as it is created, so it is handled in application space.)
    """
    # Set standard values for OpenTelemetry configuration. We don't expect to change
    # these for different pipeline environments, and they must be env vars, so they
    # are published here instead of via a .env file for example.
    for key, default_value in (
        # Upload all traces (do not downsample).
        ("OTEL_TRACES_SAMPLER", "traceidratio"),
        ("OTEL_TRACES_SAMPLER_ARG", "1.0"),
        ("OTEL_RESOURCE_ATTRIBUTES", "SampleRate=1"),
    ):
        if key not in os.environ:
            os.environ[key] = default_value

    if instrument_requests:
        RequestsInstrumentor().instrument()

    trace_provider = TracerProvider(
        # The name of what we're tracing; in Honeycomb, the dataset.
        resource=Resource(
            attributes={
                SERVICE_NAME: service_name,
            }
        ),
        # Increase the default event limit from 128 per span, since we emit logs as
        # events. Otherwise logs get lost.
        # https://opentelemetry-python.readthedocs.io/en/latest/sdk/trace.html#opentelemetry.sdk.trace.SpanLimits
        span_limits=SpanLimits(
            max_events=SpanLimits.UNSET,  # no limit (v. None which gets the default)
        ),
    )
    # Use SimpleSpanProcessor to work with Dagster's start_method:forkserver, since
    # BatchSpanProcessor is not fork-safe. See:
    # https://opentelemetry-python.readthedocs.io/en/latest/examples/fork-process-model/README.html
    # If our span volume gets too high we may want to switch back to BatchSpanProcessor,
    # if only for full pipeline runs or deployed jobs.
    processor = SimpleSpanProcessor(
        OTLPSpanExporter(
            endpoint=otlp_endpoint,
            headers=otlp_headers,
        )
    )
    trace_provider.add_span_processor(processor)
    opentelemetry.trace.set_tracer_provider(trace_provider)



"""Common string values for use as OpenTelemetry attributes.

Plain strings are used because:
*   Enums cannot be subclassed to add more values. Also iterating over all available
    attribute values or testing whether unknown strings are enum members are not
    likely use cases.
*   The aenum (Advanced Enum) library allows adding values to an existing enum, but
    then import order could affect the available values in an enum.
"""


class SpanName:
    """Names of spans which may be referenced in multiple places.

    These might be reference in application code, or in external dashboards/alerts.
    """

    #: The top-level span; a default name for the first span in the trace.
    ROOT = "trace-root"


class EventAttrKey:
    """Event attribute keys which may be referenced in external queries."""

    #: Types of events.
    TYPE = "type"
    #: The name field for a duration event. This cannot simply be "name" which is
    #: reserved for the actual event name.
    DURATION_NAME = "duration_name"
    #: The elapsed time value (in seconds) represented by a duration event.
    DURATION_SECONDS = "duration_s"


class EventAttrValue:
    """Event attribute values which may be referenced / filtered by in external queries."""

    #: An event type capturing the time taken by some portion of the pipeline.
    DURATION = "duration"
    #: An event type for log messages.
    LOG_MESSAGE = "log_message"



from typing import List


def get_run_id_and_ancestors(context) -> List[str]:
    """Returns a list starting with the current run's ID and then any ancestor runs.

    This allows searching for events when a run as been retried, and events might
    be attached to a parent run instead of this one.
    """
    run_id = context.run_id
    run_ids = [run_id]
    while True:
        run = context.instance.get_run_by_id(run_id)
        if run.parent_run_id:
            run_id = run.parent_run_id
            run_ids.append(run_id)
        else:
            break
    return run_ids



import pytest
import time
from unittest.mock import MagicMock, patch

from dagster import (
    OpExecutionContext,
    Output,
    graph,
    op,
)
import opentelemetry

from form_observability.dagster_otel.otel_ops import (
    CONFIGURE_OBSERVABILITY_RESOURCE_KEY,
    otel_op,
    publish_current_trace_context,
    _set_trace_context,
)
from form_observability.dagster_otel.testing import noop_resource


def test_publish_trace_context_logs_asset(mock_get_current_span):
    mock_context = MagicMock(spec=OpExecutionContext)
    mock_context.solid_handle.path = ["main_graph", "subgraph", "my_op"]

    publish_current_trace_context(mock_context)

    mock_context.log_event.assert_called_once()
    first_call = mock_context.log_event.call_args_list[0]
    materialization = first_call[0][0]  # first arg from the call's args list
    assert "traceparent" in materialization.metadata_entries[0].entry_data.data
    assert materialization.asset_key.path[0] == "main_graph.subgraph"


def test_otel_op_finds_trace_context_and_catches_exception():
    @op
    def root_trace_context_op(context):
        publish_current_trace_context(context)
        return 1

    # This op looks for a trace context and will error if it can't find it. It should
    # find the trace context published by the above op.
    @otel_op(
        on_exception_return=Output(5, "result"),
    )
    def test_op(context, v):
        # This exception should be caught because of the on_exception_return.
        raise RuntimeError("Catch me, one test had an error but the pipeline is OK.")

    @graph
    def otel_graph():
        test_op(root_trace_context_op())

    result = otel_graph.execute_in_process(
        resources={
            CONFIGURE_OBSERVABILITY_RESOURCE_KEY: noop_resource,
        }
    )
    assert result.success
    assert result.output_for_node("test_op") == 5


@pytest.mark.no_mock_get_current_span
def test_after_set_trace_context_trace_id_matches_passed_context():
    trace_id = 0x4A084C716BBAB165CBC4471FB5A1DEF0
    assert (
        opentelemetry.trace.get_current_span().get_span_context().trace_id != trace_id
    )
    carrier = {
        "traceparent": f"00-{trace_id:x}-d1ade6c2175f72d5-01",
    }
    _set_trace_context(carrier)
    assert (
        opentelemetry.trace.get_current_span().get_span_context().trace_id == trace_id
    )



"""OpenTelemetry tracing setup.

This module handles spans and events, but should not be aware of any implementation
under other pipeline modules (including as orchestration/ where Dagster/OTel
integration lives).

Usage:

_tracer = ContextAwareTracer(__name__)
with _tracer.start_as_current_span(
    "some work",
    attributes={
        # Attributes added to EventAttrKey/Value...
        EventAttrKey.TEST_ID: "TST22",
        # ...or just use one-off keys.
        "foo": 42,
    },
):
    # This event will have the test_id attribute, same as the parent span.
    _tracer.add_event(EventAttrValue.TEST_SKIPPED)

    # If you get more details after opening the span...
    metadata = db.get...

    # This adds cell_id as an attribute to the parent span since update_current_span=True.
    with log_ctx.set({EventAttrKey.CELL_ID: metadata["cell_id"]}, update_current_span=True):
        # Any further spans/events/logs in helper() will have test_id and cell_id.
        helper()
"""
from contextlib import contextmanager
import functools
import platform
import threading
from typing import Optional, Any, Dict

import opentelemetry

from form_observability.otel_value import (
    EventAttrKey,
    EventAttrValue,
    SpanName,
)


class _ObservabilityContext:
    """Information about the execution context, for each thread.

    This class is thread-safe and meant to be used by sharing a single instance of it
    with the whole application.

    Use the global ctx object created below.
    """

    #: Contains the isolated stacks for each thread.
    #: Each dict in a list is a new context on a stack, extending the previous context:
    #: Thread 1: [{}]
    #: Thread 2: [{}, {'a': 2}, {'a': 2, 'b': 0}]
    #: Thread 3: [{}, {'a': 3}]
    _ctx = threading.local()

    @property
    def _stack(self):
        """Lazily create and return the stack for the current thread.

        This makes sure that if the ctx is created in one thread but then used from
        another a per-thread stack is still created.
        """
        if not hasattr(self._ctx, "stack"):
            # Add these attrs globally.
            self._ctx.stack = [
                {
                    "host.name": platform.node(),
                    "os.type": platform.system().lower(),
                },
            ]
        return self._ctx.stack

    @property
    def current_ctx(self) -> dict:
        return self._stack[-1]

    @contextmanager
    def set(
        self,
        attributes: Dict[str, Any],
        update_current_span=False,
    ):
        """
        Create a new context valid for a block of code and add a key to it.

        After the block of code has ended, the previous context is restored.
        Because manually managing start and end of context is tedious, this method is
        only meant to be used with a `with` statement.

        :param attributes: Key-value pairs to add to the context.
        """
        # TODO If updating to Python 3.9, use | to update dicts more concisely.
        new_attrs = self.current_ctx.copy()
        filtered = _filter_attributes(attributes)
        new_attrs.update(filtered)
        self._stack.append(new_attrs)
        if update_current_span:
            opentelemetry.trace.get_current_span().set_attributes(filtered)
        try:
            yield
        finally:
            self._stack.pop()

    def get(self, *args, **kwargs):
        """
        Get the value for a key from the current context.
        """
        return self.current_ctx.get(*args, **kwargs)


#: Object to use throughout the application to attach context to logs and traces.
#:
#: Usage:
#:      with ctx.set({EventAttrKey.TEST_ID: "TST22"}):
#:          # test_id is added to the generated LogRecord.
#:          _log.info(f"Doing some work.")
#:          # test_id can also be read back explicitly.
#:          print("Working on test_id", ctx.get("test_id"))
ctx = _ObservabilityContext()


def _convert_types(v: Any) -> Any:
    """Converts numpy int64s to ints."""
    if type(v).__name__ in ("int8", "int32", "int64"):
        return int(v)
    return v


def _filter_attributes(d: Dict) -> Dict:
    """Pre-process span/event attributes to convert complex types and remove None values.

    Spans warn if the attribute value is None, so filter None values out. This
    filtering is not strictly necessary for span events, which accept None values.
    """
    filtered = {}
    for k, v in d.items():
        if v is None:
            continue
        filtered[_convert_types(k)] = _convert_types(v)
    return filtered


class ContextAwareTracer:
    """Main entry point for starting spans and adding events. See usage in module doc.

    Wrapper for OpenTelemetry interfaces with helpers that automatically pick up
    ctx.current_ctx and/or update it.
    """

    def __init__(self, library_name):
        self._tracer = opentelemetry.trace.get_tracer(library_name)

    def _get_required_current_span(self):
        span = opentelemetry.trace.get_current_span()
        if span is None or span.get_span_context().span_id == 0:
            raise RuntimeError(f"No current span.")
        return span

    def add_duration_event(
        self,
        duration_name: str,
        dt: float,
        attributes: Dict = {},
    ) -> None:
        """Records a duration value (duration_s) on the current span.

        :param duration_name: The value to use for the duration_name attribute on the event.
            Queries are expected to reference this value.
        :param attributes: Additional event attributes (merged with ctx.current_ctx).
        """
        if dt is None:
            raise ValueError("Bad dt value for add_duration_event: {dt}.")

        event_attrs = ctx.current_ctx.copy()
        event_attrs.update(
            {
                EventAttrKey.TYPE: EventAttrValue.DURATION,
                EventAttrKey.DURATION_NAME: duration_name,
                EventAttrKey.DURATION_SECONDS: dt,
            }
        )
        event_attrs.update(attributes)
        self._get_required_current_span().add_event(
            # Use the duration name as the event name too, for UI visibility.
            duration_name,
            attributes=_filter_attributes(event_attrs),
        )

    def add_event(self, event_type: str, attributes: Dict = {}) -> None:
        """Records an event on the current span.

        :param event_type: Used as the span name, as well as the "type" attribute.
        :param attributes: Additional event attributes (merged with ctx.current_ctx).
        """
        event_attrs = ctx.current_ctx.copy()
        event_attrs.update({EventAttrKey.TYPE: event_type})
        event_attrs.update(attributes)
        self._get_required_current_span().add_event(
            _convert_types(event_type),
            attributes=_filter_attributes(event_attrs),
        )

    def traced(self, wrapped):
        """Decorator to trace a function.

        Usage:
            _tracer = ContextAwareTracer(..)
            class Me:
                @_tracer.traced
                def my_fn(self, arg):
                    pass

        This is similar to OTel's builtin decorator, but uses the function name and
        integrates with context.
        https://opentelemetry.io/docs/instrumentation/python/manual/#creating-spans-with-decorators
        """

        @functools.wraps(wrapped)
        def wrapper(*args, **kwargs):
            with self.start_as_current_span(wrapped.__name__):
                return wrapped(*args, **kwargs)

        return wrapper

    @contextmanager
    def start_as_current_span(
        self,
        name: str,
        attributes: Dict[str, Any] = {},
        **kwargs,
    ):
        """Starts a span with the given name, merging ctx.current_ctx with attributes."""
        span_attrs = ctx.current_ctx.copy()
        span_attrs.update(attributes)
        with ctx.set(attributes or {}), self._tracer.start_as_current_span(
            _convert_types(name),
            attributes=_filter_attributes(span_attrs),
            **kwargs,
        ) as span:
            yield span



import pytest
import time
from unittest.mock import MagicMock, patch

from dagster import (
    OpExecutionContext,
    Output,
    graph,
    op,
)

from form_observability.dagster_otel.timing import (
    get_timing_since_step_started,
)
from form_observability.dagster_otel.testing import noop_resource


def test_get_timing():
    def later():
        """Returns a time 10s later than now, to simulate more time having passed."""
        return time.time() + 10

    @op
    def start_op():
        return 1

    @op
    def end_op(context, v):
        # This op should be able to find the previous op and return details about the
        # elapsed time.
        return get_timing_since_step_started(context, "start_op", seconds_fn=later)

    @graph
    def timing_graph():
        end_op(start_op())

    result = timing_graph.execute_in_process()
    assert result.success
    assert result.output_for_node("end_op").dt_s == pytest.approx(10, 0.1)



#!/usr/bin/env python3
"""Demonstrate a multithreaded program which connects child traces to a parent.

This shows how to set up simple trace context propagation.
"""
import logging
import os
import random
import threading
import time

import opentelemetry

from form_observability import (
    ContextAwareTracer,
    OtelSpanEventHandler,
    configure,
)


_trace = ContextAwareTracer(__name__)
_log = logging.getLogger(__name__)


def _thread_worker(test_id: str, otel_ctx) -> None:
    """Worker which re-attaches the OpenTelemetry trace context.

    OpenTelemetry does not automatically pick up a parent thread's trace context in a
    child thread. So this worker must be explicitly passed the value from
    context.get_current() which it can then context.attach(..). Thus the child's span
    is parented under the main span.
    """
    opentelemetry.context.attach(otel_ctx)
    sleep_s = random.random() * 4
    with _trace.start_as_current_span(test_id, attributes={"sleep_s": sleep_s}):
        time.sleep(sleep_s)
        _trace.add_event("success")


@_trace.traced
def main():
    _log.info("Preparing to process in threads.")
    otel_ctx = opentelemetry.context.get_current()
    threads = []
    for test_id in ("one", "two", "three", "four", "five", "six"):
        thread = threading.Thread(target=_thread_worker, args=(test_id, otel_ctx))
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()
    _log.info("Demo processing complete.")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    header_env = "OTEL_EXPORTER_OTLP_HEADERS"
    endpoint_env = "OTEL_EXPORTER_OTLP_ENDPOINT"
    if header_env not in os.environ:
        _log.error(f"To upload traces, you must set {header_env}, see README.md.")
    configure(
        "form_observability_example",
        otlp_headers=os.environ.get(header_env),
        otlp_endpoint=os.environ.get(endpoint_env, "https://api.honeycomb.io"),
    )
    _log.addHandler(OtelSpanEventHandler())
    main()



# Make commonly used symbols importable from form_observability.dagster_otel.
from .otel_ops import (
    CONFIGURE_OBSERVABILITY_RESOURCE_KEY,
    otel_op,
    publish_current_trace_context,
)
from .timing import TimingInfo, get_timing_since_step_started
from .util import get_run_id_and_ancestors



"""OTel tracing/logging tools, emphasizing data pipeline use cases with Dagster."""

#: Semantic versioning number. The three dotted numbers are:
#: *   Major: breaking API changes, significant feature additions.
#: *   Minor: standard feature additions and improvements. No breaking changes.
#: *   Micro: small bug fixes.
__version__ = "0.1.1"

# Make commonly used symbols importable from form_observability.
from .context_aware import ContextAwareTracer, ctx
from .log import OtelSpanEventHandler
from .otel_value import EventAttrKey, EventAttrValue, SpanName
from .pipeline_setup import configure



"""Dagster job with OpenTelemetry trace propagation.

Run with:
    dagster job execute --python-file examples/dagster_job.py
"""
import os

from dagster import (
    Output,
    job,
    op,
    resource,
)

from form_observability.dagster_otel import (
    CONFIGURE_OBSERVABILITY_RESOURCE_KEY,
    otel_op,
    publish_current_trace_context,
)
from form_observability import (
    ContextAwareTracer,
    SpanName,
    ctx,
    configure,
)


_trace = ContextAwareTracer(__name__)


@resource
def configure_observability_resource(init_context):
    """This resource sets up the span exporter (before each op executes)."""
    header_env = "OTEL_EXPORTER_OTLP_HEADERS"
    endpoint_env = "OTEL_EXPORTER_OTLP_ENDPOINT"
    if header_env not in os.environ:
        init_context.log.error(
            f"To upload traces, you must set {header_env}, see README.md."
        )
    configure(
        "form_observability_example",
        otlp_headers=os.environ.get(header_env),
        otlp_endpoint=os.environ.get(endpoint_env, "https://api.honeycomb.io"),
    )


@op(
    required_resource_keys={CONFIGURE_OBSERVABILITY_RESOURCE_KEY},
)
def root_trace_context_op(context):
    """Each job/subgraph should have one regular op to start and publish the OTel trace."""
    with _trace.start_as_current_span(SpanName.ROOT):
        publish_current_trace_context(context)
        return "TS123"


@_trace.traced
def _do_work(test_id: str):
    """This throws an exception which should be caught by the on_exception_return."""
    raise RuntimeError("Catch me, one test had an error but the pipeline is OK.")


@otel_op(
    # If anything in this op raises an exception, catch it and report it using
    # OpenTelemetry, but do not let it cause the Dagster op to fail. This is useful
    # in cases where the overall pipeline can continue running, and you have
    # monitoring in place to investigate specific errors.
    on_exception_return=Output(5, "result"),
)
def test_op(context, test_id):
    """Other ops in the graph use @otel_op to pick up that trace context.

    This op looks for a trace context and will error if it can't find it. It should
    find the trace context published by the above op.
    """
    with ctx.set({"test_id": test_id}, update_current_span=True):
        _do_work(test_id)


@job(
    # The job will use this resource to set up the span exporter. Every @otel_op
    # requires the observability resource.
    resource_defs={
        CONFIGURE_OBSERVABILITY_RESOURCE_KEY: configure_observability_resource,
    },
)
def otel_graph():
    test_op(root_trace_context_op())



import pytest
from types import SimpleNamespace
from unittest import mock

import opentelemetry
from opentelemetry.trace.span import NonRecordingSpan


@pytest.fixture(autouse=True)
def mock_get_current_span(request):
    """Provide a standin current span for unit tests.

    Since there is no active OpenTelemetry configuration in tests, there will never
    be a real span with a trace/span ID otherwise.

    Allow this autouse fixture to be disabled using
    @pytest.mark.no_mock_get_current_span, for the observability test that
    explicitly verifies trace IDs in trace propagation.
    """
    if "no_mock_get_current_span" in request.keywords:
        yield
    else:
        span = NonRecordingSpan(
            SimpleNamespace(trace_id=11, span_id=22, trace_flags=0, trace_state=False)
        )
        with mock.patch.object(
            opentelemetry.trace, "get_current_span", return_value=span
        ):
            yield



import pytest
from unittest import mock

from dagster import resource


@resource
def noop_resource():
    """A Dagster resource to fulfill cases where the resource will not be used.

    For the observablity resource when we don't want to record spans, in tests.
    """


@pytest.fixture
def skip_otel_op_trace_context():
    """Do not try to find a trace context in Dagster @otel_op s.

    This is necessary when running single ops in unit tests, because there will be no
    published trace context, and ops are also not provided handles (used in looking up
    the trace context).
    """
    with mock.patch(
        "form_observability.dagster_otel.otel_ops._find_trace_context",
        return_value={},
    ):
        yield
